// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *	Reliable Broadcast over unreliable low-bandwidth (satellite)
 *			channels with no feedback
 *
 * 			  Link Access Protocol
 *
 * Defines abstract classes that do all actual interaction with a link
 * (communication channel). The Link interface is responsible
 * for writing/identifying a start of the packet, and reading/writing
 * standard data types and arrays of bytes.
 * Note, all multi-byte data types (including integers of various sizes)
 * are read/written in the standard network order, that is, the Most
 * Significant Byte First (Big-Endian)
 *
 * $Id: LAP.h,v 1.1 1996/07/26 15:04:21 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifndef __GNUC__
#pragma once
#else
#pragma interface
#endif

#ifndef __LAP_h
#define __LAP_h

#include "crc.h"

		// Exceptions that can happen while reading from a link
struct rc_timeout {};

struct rc_io_error
{
  const int errno;
  rc_io_error(const int _errno) : errno(_errno) {}
};

class rc_input_error : public rc_io_error
{
public:
  rc_input_error(const int _errno) : rc_io_error(_errno) {}
};

class rc_file_error : public rc_io_error
{
public:
  rc_file_error(const int _errno) : rc_io_error(_errno) {}
};

struct rc_eof {};

struct rc_bad_packet
{
  enum Reason {size_error=1, CRC_error, inconsistent_size,
  	       inconsistent_seqno, bad_segment0} reason;
  rc_bad_packet(Reason _reason) : reason(_reason) {}
  operator const char * (void);
};

		// Standard Link Access Interface
class LAPIn
{
public:
  virtual unsigned char read_byte(void) = 0;
  virtual void look_for_start_of_packet(void) = 0;
  virtual void ending_packet(void) = 0;
};

class LAPOut
{
public:
  virtual void write_byte(const unsigned char a_byte) = 0;
  virtual void begin_packet(void) = 0;		// start a new packet
  virtual void finish_packet(void) = 0;		// end of the packet

};

inline LAPOut& operator << (LAPOut& lap, const CRC16 crc)
{ 
  lap.write_byte((crc.curr_crc)>>8);
  lap.write_byte(crc.curr_crc);
  return lap;
}

inline CRC16::CRC16(LAPIn& lap)
{
  curr_crc = lap.read_byte();
  curr_crc = (curr_crc<<8) | lap.read_byte();
}

#endif
