// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *	Reliable Broadcast over unreliable low-bandwidth (satellite)
 *			channels with no feedback
 *
 * 		Link Access Protocol - a Network Link
 *
 * An env variable LAPNETHOST must specify a target of a broadcast (for
 * a broadcaster) or a source of a broadcast (for a receiver):
 * a host name, or an IP address in a dot-notation. The address may be
 * a brodcast/multicast address. A receiver doesn't actually require
 * LAPNETHOST: if omitted, it is assumed "0.0.0.0"
 * (That is, the receiver would accept datagrams from anybody).
 * An env variable LAPNETPORT must specified a UDP port for this UDP
 * service.
 *
 * In addition, an env variable LAPNETERR may be specified to tell
 * the error rate to be introduced to the input stream. E.g., if
 * LAPNETERR=10, every 10th byte (on average) read from the file
 * would be distorted.
 *
 * In this particular implementation, a segment is sent/received in 
 * a single UDP packet. Although it is not very difficult to pack
 * several segments into a packet to utilize the bandwdith better.
 *
 * $Id: LAPNet.cc,v 1.2 1996/08/13 16:14:41 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifdef __GNUC__
#pragma implementation
#endif

#include "util.h"
#include "LAPNet.h"


static const char LAPNet_host [] = "LAPNETHOST";
static const char LAPNet_port [] = "LAPNETPORT";
static const char LAPError_rate_name [] = "LAPNETERR";
static const int Datagram_buffer_size = 256;

/*
 *------------------------------------------------------------------------
 *        		A part of my TCP/IP class lib
 */

#include <fstream.h>
#include <netdb.h>
//#include <fcntl.h>
//#include <sys/file.h>

extern "C"
{
                                // Convertion between the host and the network
                                // byte orders
#ifndef htons
unsigned short htons(unsigned int data);        // For a short data item
unsigned short ntohs(unsigned int data);        // For a short data item
unsigned long  htonl(unsigned long data);       // For a long data item
#endif
#include <arpa/inet.h>
}

                                // Make sure that a system call went well
                                // If a system call (say, fcntl()) fails
                                // it returns -1 and sets the errno
#define do_well(FX) (void)( (FX) < 0 ? (perror("System error"), \
        _error("Failed system call " #FX " at line %d of `%s'.\n", \
               __LINE__, __FILE__),0) : 1)


/*
 *------------------------------------------------------------------------
 *                        Class IPaddress that provides
 *                    addressing and naming in the IP network
 */


                                // Obtains the IP address of the specified
                                // host. The host name may be specified either
                                // in the dot notation, or as a host "name".
                                // Name resolution is performed in the
                                // latter case.
IPaddress::IPaddress(const char * host_name)
{
                                        // First check to see if the host
                                        // name is specified in the IP address
                                        // dot notation
  if( (address = inet_addr(host_name)) != (unsigned long)(-1) )
    return;

  struct hostent *host_ptr = ::gethostbyname(host_name);
  if( host_ptr == 0 )
    _error("Host name '%s' cannot be resolved",host_name);
  if( host_ptr->h_addrtype != AF_INET )
    _error("'%s' isn't an Internet site, or so the DNS says",host_name);

  address = *(unsigned long *)(host_ptr->h_addr);
}


                                // get a STATIC symbolic representation
                                // of an IP address
                                // We use the fact that the IP address is in
                                // the net order, that is, the MSB first
                                // We do reverse name resolution if necessary
IPaddress::operator const char * (void) const
{
  struct hostent *host_ptr = ::gethostbyaddr((char *)&address,
                                             sizeof(address), AF_INET);
  if( host_ptr != 0 )
    return host_ptr->h_name;

                                        // Reverse DNS failed, use print in
                                        // the dot notation
  static char buffer[80];
  const unsigned int native_addr = ntohl(address);
  sprintf(buffer,"%0d.%0d.%0d.%0d", (native_addr >> 24) & 0xff,
          (native_addr >> 16) & 0xff, (native_addr >> 8) & 0xff,
          native_addr & 0xff);
  return buffer;
}

/*
 *------------------------------------------------------------------------
 *                              Sockets
 */
                                // Create an IP address space entity
SocketAddr::SocketAddr(const IPaddress host, const short port_no)
{
  sin_family = AF_INET;
  sin_port = htons((short)port_no);
  sin_addr.s_addr = host.net_addr();
  assert( port_no > 0 );
}

                        // Give a STATIC string representation of a socket
                        // address
SocketAddr::operator const char * (void) const
{
  assert( sin_family == AF_INET );
  static char buffer[80];

  sprintf(buffer,"%s:%d",(const char *)IPaddress(sin_addr.s_addr),
          ntohs((short)sin_port));
  return buffer;
}

/*
 *------------------------------------------------------------------------
 *                 		DataGrams
 */

                                // Create a datagram buffer
DataGramBuffer::DataGramBuffer(const int size)
        : data_buffer(new unsigned char[size]),
          buffer_size(size)
{
  read_ptr = write_ptr = data_buffer;
  assert( buffer_size > 0 );
}


                                // Destructor
DataGramBuffer::~DataGramBuffer(void)
{
  assert( data_buffer != 0 && buffer_size > 0 );
  delete [] data_buffer;
}

				// Put a byte in the datagram buffer
void DataGramBuffer::put_char(unsigned char a_byte)
{
  assert( write_ptr < data_buffer + buffer_size );
  *write_ptr++ = a_byte;
}

				// Get a character from the buffer
unsigned char DataGramBuffer::get_char(void)
{
  if( read_ptr == write_ptr )
    throw rc_bad_packet(rc_bad_packet::size_error);
  assert( read_ptr < write_ptr );
  return *read_ptr++;
}

				// Send out the whole datagram (and empty the
				// buffer)
void DataGramBuffer::send(UDPsocketOut& out_link)
{
  assert( write_ptr > read_ptr && read_ptr >= data_buffer );
  const size_t data_size_to_send = write_ptr - read_ptr;
  assert( data_size_to_send > 0 );
  
  const size_t data_size_sent = out_link.send(read_ptr,data_size_to_send);
  if( data_size_sent != data_size_to_send )
    perror("Datagram send error"),
    _error("Only %d bytes of %d has been sent. See the error above",
           data_size_sent,data_size_to_send);
  read_ptr = write_ptr = data_buffer;
}


				// Receive a datagram, disregarding whatever
				// was in the buffer
void DataGramBuffer::receive(UDPsocketIn& in_link)
{
  read_ptr = write_ptr = data_buffer;
  const size_t data_size_received =
  		in_link.receive(write_ptr,buffer_size);
  assert( data_size_received > 0 );
  write_ptr += data_size_received;
}

/*
 *------------------------------------------------------------------------
 *                   UDPsocket - the actual transport
 */

                                // Make a generic socket with a binding
                                // to some temporary port
                                // Use it to send packets
UDPsocket::UDPsocket(void)
     : socket_handle(socket(AF_INET,SOCK_DGRAM,0))
{
  do_well( socket_handle );
}


                                // Make a socket and bind it to a
                                // host and/or a UDP port
UDPsocket::UDPsocket(const SocketAddr bind_address)
     : socket_handle(socket(AF_INET,SOCK_DGRAM,0))
{
  do_well( socket_handle );	
  do_well( ::bind(socket_handle, (sockaddr *)bind_address,
                  sizeof(bind_address)) );
}

UDPsocket::~UDPsocket(void)
{
  assert( socket_handle >= 0 );
  assert( ::close(socket_handle) == 0 );
}

                                // Create a transmitting socket
                                // Note the destination address should
                                // be valid (or specified as a broadcast
                                // address with 0 as a wildcard byte).
                                // Otherwise, no packet is sent, but no
                                // error is reported either.
UDPsocketOut::UDPsocketOut(const SocketAddr _dest_addr)
	: dest_addr(_dest_addr)
{
//  char opt = 255;
//  do_well( setsockopt(socket_handle,SOL_SOCKET,SO_BROADCAST,&opt,sizeof(opt)) );
}

                                // Send a datagram
size_t UDPsocketOut::send(const unsigned char * const buffer,
			  const size_t length)
{
  assert( length > 0 );
  
  const int data_size_sent = ::sendto(socket_handle,(const char *)buffer,
  					length,0,
                              		(sockaddr *)dest_addr,sizeof(dest_addr));
  if( data_size_sent < 0 )
    perror("Datagram send error"),
    _error("Failed to send a datagram due to the error above");
  return data_size_sent;
}


                                // Create a receiving socket, accepting
                                // connectionas from dest_addr
                                // The host part of dest_addr may be a
                                // multicast or a broadcast addr, or even
                                // 0.
UDPsocketIn::UDPsocketIn(const SocketAddr dest_addr)
	: UDPsocket(dest_addr)
{
}

                                // Receive a datagram into a specified buffer
size_t UDPsocketIn::receive(unsigned char * const buffer, const size_t length)
{
  int fromlen = sizeof(src_addr);
  assert( length > 0 );
  const int received_data_len = recvfrom(socket_handle,(char *)buffer,length,0,
                               (sockaddr *)src_addr, &fromlen);
  if( received_data_len <= 0 )
    throw rc_input_error(errno);
  return received_data_len;
}

/*
 *------------------------------------------------------------------------
 *         		Link Access interface
 */

static SocketAddr get_netaddr(const char * title,
			      const char * default_ip_addr)
{
  IPaddress in_addr(xgetenv(LAPNet_host,default_ip_addr));
  SocketAddr sock_addr(in_addr,atoi(xgetenv(LAPNet_port,0)));
  Logger::log("Net Link: %s %s",title,(const char *)sock_addr);
  return sock_addr;
}

			// Open the net access link and initialize the
			// data buffer
LAPNetOut::LAPNetOut(void)
	: out_link(get_netaddr("UDP datagrams would be sent",0)),
	  buffer(Datagram_buffer_size)
{
  Logger::log("UDP datagrams would be sent in chunks of at most %d bytes",
  	      Datagram_buffer_size);
}

			// Done with the net
LAPNetOut::~LAPNetOut(void)
{
}

  			// The standard LAPOut interface
void LAPNetOut::write_byte(const unsigned char a_byte)
{
  buffer.put_char(a_byte);
}

void LAPNetOut::begin_packet(void)
{
  write_byte(SOH);
}

void LAPNetOut::finish_packet(void)
{
  buffer.send(out_link);
}


static int get_error_rate(void)
{
  const int rate = atoi(xgetenv(LAPError_rate_name,"0"));
  assert( rate >= 0 );
  if( rate != 0 )
    Logger::log("every %d-th byte (on average) read from the file would be "
    		"mangled",rate);
  return rate == 0 ? Random::One_in_n::never : rate;
}


			// Open the net access link and initialize the
			// data buffer
LAPNetIn::LAPNetIn(void)
	: in_link(get_netaddr("expecting UDP datagrams from","0.0.0.0")),
	  buffer(Datagram_buffer_size),
	  error_event(get_error_rate())
{
  Logger::log("UDP datagrams are not expected to be bigger than %d bytes",
   	      Datagram_buffer_size);
}

			// Done with the net
LAPNetIn::~LAPNetIn(void)
{
}

			// (unreliable) read from the datagram
int LAPNetIn::get_char(void)
{
  const int c = buffer.get_char();
  if( (bool)error_event )
    return c ^ 1;		// mangle the byte; toggle the lowest bit
  else
    return c;
}

  			// The standard LAPIn interface
unsigned char LAPNetIn::read_byte(void)
{
  const int c = get_char();
  if( c == EOF )
    throw rc_input_error(errno);
  return c;
}

void LAPNetIn::look_for_start_of_packet(void)
{
  register int c;
  while( (buffer.receive(in_link), c=get_char()) != LAPNetOut::SOH )
    ;
}

void LAPNetIn::ending_packet(void)
{
}
