// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *	Reliable Broadcast over unreliable low-bandwidth (satellite)
 *			channels with no feedback
 *
 * 		Link Access Protocol - a Network Link
 *
 * This is a particular type of link access via a TCP/IP network using
 * UDP datagrams. Of all the protocols of the TCP/IP suite, UDP is best
 * suited for a packet-oriented broadcast without any feedback.
 * 
 * An env variable LAPNETHOST must specify a target of a broadcast (for
 * a broadcaster) or a source of a broadcast (for a receiver):
 * a host name, or an IP address in a dot-notation. The address may be
 * a brodcast/multicast address. A receiver doesn't actually require
 * LAPNETHOST: if omitted, it is assumed "0.0.0.0"
 * (That is, the receiver would accept datagrams from anybody).
 * An env variable LAPNETPORT must specified a UDP port for this UDP
 * service.
 *
 * $Id: LAPNet.h,v 1.1 1996/08/01 17:00:56 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifndef __GNUC__
#pragma once
#else
#pragma interface
#endif

#ifndef __LAPNet_h
#define __LAPNet_h

#include "LAP.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

                    // Networking classes
class IPaddress
{
  friend class SocketAddr;

  unsigned long address;                // Address: 4 bytes in the network
                                        // byte order
  IPaddress(const unsigned int netaddr) : address(netaddr) {}

public:
  IPaddress(void)   { address = INADDR_ANY; }   // Wildcard address
  IPaddress(const char * name);         // Involves the name resolution
  unsigned long net_addr(void) const { return address; }
  operator const char * (void) const;   // get a STATIC symbolic representation
                                        // of an IP address
};

class SocketAddr : sockaddr_in
{
  friend class UDPsocket;
  friend class UDPsocketIn;
  SocketAddr(void) {}

public:
  SocketAddr(const IPaddress host, const short port_no);
  operator sockaddr * (void) const      { return (sockaddr *)this; }
  operator const char * (void) const;   // Give a STATIC string representation
};

class UDPsocket
{
protected:
  const int socket_handle;              // Socket handle = file handle

public:
  UDPsocket(void);                      // Make an implicitly bound socket
  					// (useful for sending out datagrams)
                                        // Bind the socket explicitly
  UDPsocket(const SocketAddr addr);	// (to accept datagrams)
  ~UDPsocket(void);
};

class UDPsocketOut : public UDPsocket
{
  SocketAddr dest_addr;
public:
  UDPsocketOut(const SocketAddr _dest_addr);
  size_t send(const unsigned char * const buffer, const size_t length);
};

class UDPsocketIn : public UDPsocket
{
  SocketAddr src_addr;			// The origin of the last received packet
public:
  UDPsocketIn(const SocketAddr dest_addr);
  size_t receive(unsigned char * const buffer, const size_t length);
  const SocketAddr& q_origin(void) const	{ return src_addr; }
};

class DataGramBuffer
{
  unsigned char * const data_buffer;	// Ptr to the actual datagram buffer
  const int buffer_size;
  const unsigned char * read_ptr;
  unsigned char * write_ptr;		// Pointers for accessing the buffer

  DataGramBuffer& operator = (const DataGramBuffer&); // unimplemented and forbidden
  DataGramBuffer(const DataGramBuffer&);	      // unimplemented and forbidden

public:

  DataGramBuffer(const int _size);	// Allocate the data buffer
  ~DataGramBuffer(void);
  void put_char(unsigned char a_byte);
  unsigned char get_char(void);
  int length(void) const		{ return write_ptr - read_ptr; }
  void send(UDPsocketOut& out_link);
  void receive(UDPsocketIn& in_link);
};



class LAPNetOut : public LAPOut
{
  UDPsocketOut out_link;		// An outbound link
  DataGramBuffer buffer;

public:
  LAPNetOut(void);
  virtual ~LAPNetOut(void);
  
  enum { SOH = 'A' };		// Start of the packet char

  				// Standard LAPOut interface
  virtual void write_byte(const unsigned char a_byte);
  virtual void begin_packet(void);
  virtual void finish_packet(void);
};

class LAPNetIn : public LAPIn
{
  UDPsocketIn in_link;		// An inbound link
  DataGramBuffer buffer;
  Random::One_in_n error_event;
  int get_char(void);
  
public:
  LAPNetIn(void);
  virtual ~LAPNetIn(void);
  				// Standard LAPIn interface
  virtual unsigned char read_byte(void);
  virtual void look_for_start_of_packet(void);
  virtual void ending_packet(void);
};
#endif
