// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *	Reliable Broadcast over unreliable low-bandwidth (satellite)
 *			channels with no feedback
 *
 * 	Standard utility services used by many parts of the project
 *
 * One of them is the logging facility, which (filters) and logs various
 * system activities onto stderr, unless an environment variable 
 * BC_LOG_FILE names some other file.
 *
 * $Id: util.h,v 1.1 1996/07/26 16:17:52 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifndef __GNUC__
#pragma once
#else
#pragma interface
#endif

#ifndef __util_h
#define __util_h

#include "myenv.h"
#include "std.h"

			// Time-stamp and log a message

class Logger
{
  FILE * log_fp;		// where to log
  
  static Logger me;		// The only instance of the logger that can
  				// ever be made, and it is private.
  Logger(void);			// And this will guarantee that
				// These are deliberately private and
				// unimplemented to discourage
				// copy-constructing
  void operator = (const Logger&);
  Logger(const Logger&);

public:
  ~Logger(void);
				// Log a message (possibly followed by
				// arguments, in a manner similar to
				// printf())
  static void log(const char * message,...);
};


			// Handle a SIG from the OS
			// This is an abstract class, you have to actually
			// implement function handle() that is called when
			// the signal is received. Refrain from throwing
			// exceptions from it so the OS SIG sender could
			// get the control back.
			// The function handle() should return 'true' if
			// it handled the signal, or 'false' if it couldn't
			// (and other handlers down the chain should be
			// called - not implemented at present)
class Signal
{
  typedef void (*Disposition)(int);

  static void disposition_relay(int signal_id);
  static Signal * curr_Signal;
  
  const int signal_id;
  const Disposition prev_disposition;
  
  virtual bool handle(void) = 0;

public:
  Signal(const int _signal_id);
  virtual ~Signal(void);
  int q_signal_id(void) const	{ return signal_id; }
};


			// Random number generating service
class Random
{

  static Random me;		// The only instance of this class that can
  				// ever be made, and it is private.
  Random(void);			// And this will guarantee that

				// These are deliberately private and
				// unimplemented to discourage
				// copy-constructing
  void operator = (const Random&);
  Random(const Random&);

public:
				// Generate a random id, a positive short
				// number
  static inline short get_id(void)	{ return rand() & 0x7fff; }
  
  
  				// Generate a random event that occurs
  				// randomly uniformly once every n
  				// tries, on average
  				// Given a random number r uniformly
  				// distributed within (0,1), an event happens
  				// if r < 1/n
  				// Note that rand() is uniformly distributed
  				// over [0,2^15)
  class One_in_n
  {
    const int threshold;
  public:
   One_in_n(const int n) : threshold((1<<15)/n) { }
   inline operator bool (void)	{ return (rand() & 0x7fff) < threshold; }
    enum { never = 1<<16 };	// Value of n above to guarantee an event
    				// never occurs
  };
};


#endif
